import logging
from typing import Generic

import aiohttp_jinja2
import jwt
from aiohttp import web

from ai.backend.logging import BraceStyleAdapter
from ai.backend.wsproxy.defs import RootContext
from ai.backend.wsproxy.exceptions import BackendError, InvalidCredentials
from ai.backend.wsproxy.proxy.backend.http import HTTPBackend
from ai.backend.wsproxy.types import (
    PERMIT_COOKIE_NAME,
    Circuit,
    InferenceAppInfo,
    InteractiveAppInfo,
    RouteInfo,
    TCircuitKey,
    WebRequestHandler,
)
from ai.backend.wsproxy.utils import ensure_json_serializable, is_permit_valid, mime_match

from ..abc import AbstractFrontend

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class AbstractHTTPFrontend(Generic[TCircuitKey], AbstractFrontend[HTTPBackend, TCircuitKey]):
    root_context: RootContext

    def ensure_credential(self, request: web.Request, circuit: Circuit) -> None:
        if circuit.open_to_public:
            return

        match circuit.app_info:
            case InteractiveAppInfo():
                permit_hash = request.cookies.get(PERMIT_COOKIE_NAME)
                if not permit_hash:
                    raise InvalidCredentials("E20004: Authorization cookie not provided")
                if not is_permit_valid(
                    self.root_context.local_config.wsproxy.permit_hash_key,
                    circuit.app_info.user_id,
                    permit_hash,
                ):
                    raise InvalidCredentials("E20005: Invalid authorization cookie")
            case InferenceAppInfo():
                auth_header = request.headers.get("Authorization")
                if not auth_header:
                    raise InvalidCredentials("E20006: Authorization header not provided")
                auth_type, auth_key = auth_header.split(" ", maxsplit=2)
                if auth_type in ("BackendAI", "Bearer"):
                    token = auth_key
                else:
                    raise InvalidCredentials(
                        f"E20007: Unsupported authorization method {auth_type}"
                    )

                try:
                    decoded = jwt.decode(
                        token,
                        key=self.root_context.local_config.wsproxy.jwt_encrypt_key,
                        algorithms=["HS256"],
                    )
                except jwt.PyJWTError as e:
                    raise InvalidCredentials from e

                if decoded.get("id") != str(circuit.id):
                    raise InvalidCredentials("E20008: Authorization token mismatch")

    async def initialize_backend(self, circuit: Circuit, routes: list[RouteInfo]) -> HTTPBackend:
        return HTTPBackend(routes, self.root_context, circuit)

    async def update_backend(self, backend: HTTPBackend, routes: list[RouteInfo]) -> HTTPBackend:
        backend.routes = routes
        return backend

    async def terminate_backend(self, backend: HTTPBackend) -> None:
        return

    async def proxy(self, request: web.Request) -> web.StreamResponse | web.WebSocketResponse:
        backend: HTTPBackend = request["backend"]

        if (
            "upgrade" in request.headers.get("connection", "").lower()
            and request.headers.get("upgrade", "").lower() == "websocket"
        ):
            return await backend.proxy_ws(request)
        else:
            return await backend.proxy_http(request)

    @web.middleware
    async def exception_middleware(
        self, request: web.Request, handler: WebRequestHandler
    ) -> web.StreamResponse:
        try:
            resp = await handler(request)
        except BackendError as ex:
            if ex.status_code == 500:
                log.exception("Internal server error raised inside handlers")
            if mime_match(
                request.headers.get("accept", "text/html"), "application/json", strict=True
            ):
                return web.json_response(
                    ensure_json_serializable(ex.body_dict),
                    status=ex.status_code,
                )
            else:
                return aiohttp_jinja2.render_template(
                    "error.jinja2",
                    request,
                    ex.body_dict,
                )
        return resp

    @web.middleware
    async def cors_middleware(
        self, request: web.Request, handler: WebRequestHandler
    ) -> web.StreamResponse:
        resp = await handler(request)
        resp.headers["Access-Control-Allow-Origin"] = "*"
        return resp



import asyncio
import logging
import random
import socket
from typing import Final

from ai.backend.logging import BraceStyleAdapter
from ai.backend.wsproxy.exceptions import WorkerNotAvailable
from ai.backend.wsproxy.types import RouteInfo

from .abc import AbstractBackend

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

MAX_BUFFER_SIZE: Final[int] = 1 * 1024 * 1024


class TCPBackend(AbstractBackend):
    routes: list[RouteInfo]

    def __init__(self, routes: list[RouteInfo], *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.routes = routes

    @property
    def selected_route(self) -> RouteInfo:
        if len(self.routes) == 0:
            raise WorkerNotAvailable
        elif len(self.routes) == 1:
            selected_route = self.routes[0]
            if selected_route.traffic_ratio == 0:
                raise WorkerNotAvailable
        else:
            ratios: list[float] = [r.traffic_ratio for r in self.routes]
            selected_route = random.choices(self.routes, weights=ratios, k=1)[0]
        return selected_route

    async def bind(
        self, down_reader: asyncio.StreamReader, down_writer: asyncio.StreamWriter
    ) -> None:
        stop_event = asyncio.Event()

        async def _pipe(
            reader: asyncio.StreamReader,
            writer: asyncio.StreamWriter,
            tag="(unknown)",
        ) -> None:
            try:
                while True:
                    data = await reader.read(n=MAX_BUFFER_SIZE)
                    if not data:
                        break
                    writer.write(data)
                    await writer.drain()
                    log.debug("TCPBackend._pipe(t: {}): sent {} bytes", tag, len(data))
            except ConnectionResetError:
                log.debug("Conn reset")
                pass
            except Exception:
                log.exception("")
                raise
            finally:
                log.debug("setting stop event")
                stop_event.set()

        route = self.selected_route
        log.debug(
            "Proxying TCP Request to {}:{}",
            route.kernel_host,
            route.kernel_port,
        )

        try:
            sock = socket.socket(family=socket.AF_INET, type=socket.SOCK_STREAM)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            # unlike .frontend.tcp this has a chance of being a blocking call since kernel host can be a domain
            await asyncio.get_running_loop().run_in_executor(
                None, sock.connect, (route.kernel_host, route.kernel_port)
            )

            up_reader, up_writer = await asyncio.open_connection(sock=sock)
            log.debug(
                "Connected to {}:{}",
                route.kernel_host,
                route.kernel_port,
            )
            async with asyncio.TaskGroup() as group:
                group.create_task(_pipe(up_reader, down_writer, tag="up->down"))
                group.create_task(_pipe(down_reader, up_writer, tag="down->up"))
        finally:
            log.debug("tasks ended")
            down_writer.close()
            await down_writer.wait_closed()
        log.debug("TCP connection closed")



import base64
import enum
import hashlib
import hmac
from datetime import datetime
from pathlib import Path
from typing import Any
from uuid import UUID

import humps
from pydantic import BaseModel

from ai.backend.common.types import HostPortPair


def ensure_json_serializable(o: Any) -> Any:
    match o:
        case dict():
            return {ensure_json_serializable(k): ensure_json_serializable(v) for k, v in o.items()}
        case list():
            return [ensure_json_serializable(x) for x in o]
        case UUID():
            return str(o)
        case HostPortPair():
            return {"host": o.host, "port": o.port}
        case Path():
            return o.as_posix()
        case BaseModel():
            return ensure_json_serializable(o.model_dump())
        case enum.Enum():
            return o.value
        case datetime():
            return o.timestamp()
        case _:
            return o


def config_key_to_kebab_case(o: Any) -> Any:
    match o:
        case dict():
            return {humps.kebabize(k): config_key_to_kebab_case(v) for k, v in o.items()}
        case list():
            return [config_key_to_kebab_case(i) for i in o]
        case _:
            return o


def mime_match(base_array: str, compare: str, strict=False) -> bool:
    """
    Checks if `base_array` MIME string contains `compare` MIME type.

    :param: base_array: Array of MIME strings to be compared, concatenated with comma (,) delimiter.
    :param: compare: MIME string to compare.
    :param: strict: If set to True, do not allow wildcard on source MIME type.
    """
    for base in base_array.split(","):
        _base, _, _ = base.partition(";")
        base_left, _, base_right = _base.partition("/")
        compare_left, compare_right = compare.split(";")[0].split("/")
        if (
            not strict
            and (
                (base_left == "*" and base_right == "*")
                or (base_left == compare_left and base_right == "*")
            )
        ) or (base_left == compare_left and base_right == compare_right):
            return True
    return False


def calculate_permit_hash(hash_key: str, user_id: UUID) -> str:
    hash = hmac.new(hash_key.encode(), str(user_id).encode("utf-8"), getattr(hashlib, "sha256"))
    return base64.b64encode(hash.hexdigest().encode()).decode()


def is_permit_valid(hash_key: str, user_id: UUID, hash: str) -> bool:
    valid_hash = calculate_permit_hash(hash_key, user_id)
    return valid_hash == hash



"""
This module defines a series of Backend.AI-specific errors based on HTTP Error
classes from aiohttp.
Raising a BackendError is automatically mapped to a corresponding HTTP error
response with RFC7807-style JSON-encoded description in its response body.

In the client side, you should use "type" field in the body to distinguish
canonical error types beacuse "title" field may change due to localization and
future UX improvements.
"""

from __future__ import annotations

import json
from typing import Any, Optional

from aiohttp import web

from ai.backend.common.plugin.hook import HookResult


class BackendError(web.HTTPError):
    """
    An RFC-7807 error class as a drop-in replacement of the original
    aiohttp.web.HTTPError subclasses.
    """

    error_type: str = "https://api.backend.ai/probs/general-error"
    error_title: str = "General Backend API Error."

    content_type: str
    extra_msg: Optional[str]

    body_dict: dict[str, Any]

    def __init__(self, extra_msg: str | None = None, extra_data: Optional[Any] = None, **kwargs):
        super().__init__(**kwargs)
        self.args = (self.status_code, self.reason, self.error_type)
        self.empty_body = False
        self.content_type = "application/problem+json"
        self.extra_msg = extra_msg
        self.extra_data = extra_data
        body = {
            "type": self.error_type,
            "title": self.error_title,
        }
        if extra_msg is not None:
            body["msg"] = extra_msg
        if extra_data is not None:
            body["data"] = extra_data
        self.body_dict = body
        self.body = json.dumps(body).encode()

    def __str__(self):
        lines = []
        if self.extra_msg:
            lines.append(f"{self.error_title} ({self.extra_msg})")
        else:
            lines.append(self.error_title)
        if self.extra_data:
            lines.append(" -> extra_data: " + repr(self.extra_data))
        return "\n".join(lines)

    def __repr__(self):
        lines = []
        if self.extra_msg:
            lines.append(
                f"<{type(self).__name__}({self.status}): {self.error_title} ({self.extra_msg})>"
            )
        else:
            lines.append(f"<{type(self).__name__}({self.status}): {self.error_title}>")
        if self.extra_data:
            lines.append(" -> extra_data: " + repr(self.extra_data))
        return "\n".join(lines)

    def __reduce__(self):
        return (
            type(self),
            (),  # empty the constructor args to make unpickler to use
            # only the exact current state in __dict__
            self.__dict__,
        )


class URLNotFound(BackendError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/url-not-found"
    error_title = "Unknown URL path."


class ObjectNotFound(BackendError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/object-not-found"
    object_name = "object"

    def __init__(
        self,
        *,
        extra_msg: str | None = None,
        extra_data: Optional[Any] = None,
        object_name: str | None = None,
        **kwargs,
    ) -> None:
        if object_name:
            self.object_name = object_name
        self.error_title = f"E00002: No such {self.object_name}."
        super().__init__(extra_msg, extra_data, **kwargs)


class GenericBadRequest(BackendError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/generic-bad-request"
    error_title = "Bad request."


class RejectedByHook(BackendError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/rejected-by-hook"
    error_title = "Operation rejected by a hook plugin."

    @classmethod
    def from_hook_result(cls, result: HookResult) -> RejectedByHook:
        return cls(
            extra_msg=result.reason,
            extra_data={
                "plugins": result.src_plugin,
            },
        )


class InvalidCredentials(BackendError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/invalid-credentials"
    error_title = "Authentication credentials not valid."


class GenericForbidden(BackendError, web.HTTPForbidden):
    error_type = "https://api.backend.ai/probs/generic-forbidden"
    error_title = "Forbidden operation."


class InsufficientPrivilege(BackendError, web.HTTPForbidden):
    error_type = "https://api.backend.ai/probs/insufficient-privilege"
    error_title = "Insufficient privilege."


class MethodNotAllowed(BackendError, web.HTTPMethodNotAllowed):
    error_type = "https://api.backend.ai/probs/method-not-allowed"
    error_title = "HTTP Method Not Allowed."


class InternalServerError(BackendError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/internal-server-error"
    error_title = "Internal server error."


class ServerMisconfiguredError(BackendError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/server-misconfigured"
    error_title = "E00001: Service misconfigured."


class ServiceUnavailable(BackendError, web.HTTPServiceUnavailable):
    error_type = "https://api.backend.ai/probs/service-unavailable"
    error_title = "Serivce unavailable."


class QueryNotImplemented(BackendError, web.HTTPServiceUnavailable):
    error_type = "https://api.backend.ai/probs/not-implemented"
    error_title = "This API query is not implemented."


class InvalidAuthParameters(BackendError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/invalid-auth-params"
    error_title = "Missing or invalid authorization parameters."


class AuthorizationFailed(BackendError, web.HTTPUnauthorized):
    error_type = "https://api.backend.ai/probs/auth-failed"
    error_title = "Credential/signature mismatch."


class PasswordExpired(BackendError, web.HTTPUnauthorized):
    error_type = "https://api.backend.ai/probs/password-expired"
    error_title = "Password has expired."


class InvalidAPIParameters(BackendError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/invalid-api-params"
    error_title = "Missing or invalid API parameters."


class GraphQLError(BackendError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/graphql-error"
    error_title = "GraphQL-generated error."


class WorkerNotAvailable(BackendError):
    error_title = "Worker not available"


class PortNotAvailable(BackendError):
    error_title = "Designated port already occupied"


class UnsupportedProtocol(BackendError):
    error_title = "Unsupported protocol"


class DatabaseError(BackendError):
    error_title = "error while communicating with database"


class ContainerConnectionRefused(BackendError):
    error_title: str = "Cannot connect to Backend.AI kernel."



import importlib.resources
import logging

import aiohttp_jinja2
import jinja2
from aiohttp import web
from aiohttp.typedefs import Handler

from ai.backend.logging import BraceStyleAdapter
from ai.backend.wsproxy.exceptions import GenericBadRequest
from ai.backend.wsproxy.types import Circuit

from .abc import AbstractHTTPFrontend

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class PortFrontend(AbstractHTTPFrontend[int]):
    sites: list[web.TCPSite]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.sites = []

    async def start(self) -> None:
        config = self.root_context.local_config.wsproxy
        port_start, port_end = config.bind_proxy_port_range
        for port in range(port_start, port_end + 1):
            app = web.Application()
            app["port"] = port
            app.middlewares.extend([
                self.cors_middleware,
                self.exception_middleware,
                self._ensure_slot,
            ])
            app.router.add_route("*", "/{path:.*$}", self.proxy)

            with importlib.resources.as_file(importlib.resources.files("ai.backend.wsproxy")) as f:
                template_path = f / "templates"
                aiohttp_jinja2.setup(app, loader=jinja2.FileSystemLoader(template_path))

            runner = web.AppRunner(app, keepalive_timeout=30.0)
            await runner.setup()
            service_host = config.bind_host
            site = web.TCPSite(
                runner,
                service_host,
                port,
                backlog=1024,
                reuse_port=True,
            )
            await site.start()
            self.sites.append(site)
        log.info(
            "accepting proxy requests from {}:{}~{}",
            config.bind_host,
            port_start,
            port_end,
        )

    async def stop(self) -> None:
        for site in self.sites:
            await site.stop()

    @web.middleware
    async def _ensure_slot(self, request: web.Request, handler: Handler) -> web.StreamResponse:
        port: int = request.app["port"]

        if port not in self.circuits:
            raise GenericBadRequest(f"Unregistered slot {port}")

        circuit = self.circuits[port]
        if not circuit:
            raise GenericBadRequest(f"Circuit for registered port {port} is not available.")

        self.ensure_credential(request, circuit)
        circuit = self.circuits[port]
        backend = self.backends[port]
        request["circuit"] = circuit
        request["backend"] = backend
        return await handler(request)

    def get_circuit_key(self, circuit: Circuit) -> int:
        return circuit.port



from typing import Annotated
from uuid import UUID

from pydantic import BaseModel, Field

from ..types import (
    Circuit,
    FrontendMode,
    SessionConfig,
)


class SlotModel(BaseModel):
    frontend_mode: FrontendMode
    in_use: bool
    port: Annotated[int | None, Field(default=None)]
    subdomain: Annotated[str | None, Field(default=None)]
    circuit_id: Annotated[UUID | None, Field(default=None)]
    """
    ID of circuit slot is hosting.
    """


class StubResponseModel(BaseModel):
    success: Annotated[bool, Field(default=True)]


class CircuitListResponseModel(BaseModel):
    circuits: list[Circuit]


class ConfRequestModel(BaseModel):
    login_session_token: str | None
    kernel_host: str
    kernel_port: int
    session: SessionConfig



import datetime
import uuid
from typing import Any

from .defs import RootContext
from .exceptions import ServiceUnavailable
from .types import (
    AppMode,
    Circuit,
    EndpointConfig,
    FrontendMode,
    ProxyProtocol,
    RouteInfo,
    SessionConfig,
)


async def add_circuit(
    root_ctx: RootContext,
    session_info: SessionConfig,
    endpoint_info: EndpointConfig | None,
    app: str,
    protocol: ProxyProtocol,
    mode: AppMode,
    routes: list[RouteInfo],
    *,
    envs: dict[str, Any] = {},
    args: str | None = None,
    open_to_public=False,
    allowed_client_ips: str | None = None,
) -> Circuit:
    port_range_start, port_range_end = root_ctx.local_config.wsproxy.bind_proxy_port_range
    for port in range(port_range_start, port_range_end + 1):
        if port not in root_ctx.proxy_frontend.circuits:
            break
    else:
        raise ServiceUnavailable("Port pool exhausted")

    circuit = Circuit(
        id=uuid.uuid4(),
        app=app,
        protocol=protocol,
        worker=uuid.UUID("00000000-0000-0000-0000-000000000000"),
        app_mode=mode,
        frontend_mode=FrontendMode.PORT,
        envs=envs,
        arguments=args,
        port=port,
        user_id=session_info.user_uuid,
        access_key=session_info.access_key,
        endpoint_id=(endpoint_info.id if endpoint_info else None),
        route_info=routes,
        session_ids=[r.session_id for r in routes],
        created_at=datetime.datetime.now(),
        updated_at=datetime.datetime.now(),
        open_to_public=open_to_public,
        allowed_client_ips=allowed_client_ips,
    )
    await root_ctx.proxy_frontend.register_circuit(circuit, routes)
    return circuit



import enum
import os
import pwd
import socket
import sys
import types
import typing
from dataclasses import dataclass
from pathlib import Path
from pprint import pformat
from typing import Annotated, Any

import click
from pydantic import (
    BaseModel,
    ByteSize,
    ConfigDict,
    Field,
    GetCoreSchemaHandler,
    GetJsonSchemaHandler,
    ValidationError,
)
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import PydanticUndefined, core_schema

from ai.backend.common import config
from ai.backend.logging import LogFormat, LogLevel

from .types import EventLoopType, ProxyProtocol

_file_perm = (Path(__file__).parent / "server.py").stat()


class BaseSchema(BaseModel):
    model_config = ConfigDict(
        validate_by_name=True,
        from_attributes=True,
        use_enum_values=True,
    )


class HostPortPair(BaseSchema):
    host: Annotated[str, Field(examples=["127.0.0.1"])]
    port: Annotated[int, Field(gt=0, lt=65536, examples=[8201])]

    def __repr__(self) -> str:
        return f"{self.host}:{self.port}"

    def __str__(self) -> str:
        return self.__repr__()

    def __getitem__(self, *args) -> int | str:
        if args[0] == 0:
            return self.host
        elif args[0] == 1:
            return self.port
        else:
            raise KeyError(*args)


@dataclass
class UserID:
    default_uid: int | None = None

    @classmethod
    def uid_validator(
        cls,
        value: int | str | None,
    ) -> int:
        if value is None:
            assert cls.default_uid, "value is None but default_uid not provided"
            return cls.default_uid
        assert isinstance(value, (int, str)), "value must be an integer"
        match value:
            case int():
                if value == -1:
                    return os.getuid()
                else:
                    return value
            case str():
                try:
                    _value = int(value)
                    if _value == -1:
                        return os.getuid()
                    else:
                        return _value
                except ValueError:
                    try:
                        return pwd.getpwnam(value).pw_uid
                    except KeyError:
                        assert False, f"no such user {value} in system"

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        schema = core_schema.chain_schema([
            core_schema.union_schema([
                core_schema.int_schema(),
                core_schema.str_schema(),
            ]),
            core_schema.no_info_plain_validator_function(cls.uid_validator),
        ])

        return core_schema.json_or_python_schema(
            json_schema=schema,
            python_schema=core_schema.union_schema([
                # check if it's an instance first before doing any further work
                core_schema.union_schema([
                    core_schema.is_instance_schema(int),
                    core_schema.is_instance_schema(str),
                ]),
                schema,
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(int),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, _core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        # Use the same schema that would be used for `int`
        return handler(
            core_schema.union_schema([
                core_schema.int_schema(),
                core_schema.str_schema(),
            ])
        )


@dataclass
class GroupID:
    default_gid: int | None = None

    @classmethod
    def uid_validator(
        cls,
        value: int | str | None,
    ) -> int:
        if value is None:
            assert cls.default_gid, "value is None but default_gid not provided"
        assert isinstance(value, (int, str)), "value must be an integer"
        match value:
            case int():
                if value == -1:
                    return os.getgid()
                else:
                    return value
            case str():
                try:
                    _value = int(value)
                    if _value == -1:
                        return os.getgid()
                    else:
                        return _value
                except ValueError:
                    try:
                        return pwd.getpwnam(value).pw_gid
                    except KeyError:
                        assert False, f"no such user {value} in system"

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        schema = core_schema.chain_schema([
            core_schema.union_schema([
                core_schema.int_schema(),
                core_schema.str_schema(),
            ]),
            core_schema.no_info_plain_validator_function(cls.uid_validator),
        ])

        return core_schema.json_or_python_schema(
            json_schema=schema,
            python_schema=core_schema.union_schema([
                # check if it's an instance first before doing any further work
                core_schema.union_schema([
                    core_schema.is_instance_schema(int),
                    core_schema.is_instance_schema(str),
                ]),
                schema,
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(int),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, _core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        # Use the same schema that would be used for `int`
        return handler(
            core_schema.union_schema([
                core_schema.int_schema(),
                core_schema.str_schema(),
            ])
        )


class LogDriver(str, enum.Enum):
    CONSOLE = "console"
    LOGSTASH = "logstash"
    FILE = "file"
    GRAYLOG = "graylog"


class LogstashProtocol(str, enum.Enum):
    ZMQ_PUSH = "zmq.push"
    ZMQ_PUB = "zmq.pub"
    TCP = "tcp"
    UDP = "udp"


default_pkg_ns = {"": "WARNING", "ai.backend": "DEBUG", "tests": "DEBUG", "aiohttp": "INFO"}


class ConsoleLogConfig(BaseSchema):
    colored: Annotated[
        bool | None, Field(default=None, description="Opt to print colorized log.", examples=[True])
    ]
    format: Annotated[
        LogFormat, Field(default=LogFormat.VERBOSE, description="Determine verbosity of log.")
    ]


class FileLogConfig(BaseSchema):
    path: Annotated[Path, Field(description="Path to store log.", examples=["/var/log/backend.ai"])]
    filename: Annotated[str, Field(description="Log file name.", examples=["wsproxy.log"])]
    backup_count: Annotated[
        int, Field(description="Number of outdated log files to retain.", default=5)
    ]
    rotation_size: Annotated[
        ByteSize, Field(description="Maximum size for a single log file.", default=10 * 1024 * 1024)
    ]
    format: Annotated[
        LogFormat, Field(default=LogFormat.VERBOSE, description="Determine verbosity of log.")
    ]


class LogstashConfig(BaseSchema):
    endpoint: Annotated[
        HostPortPair,
        Field(
            description="Connection information of logstash node.",
            examples=[HostPortPair(host="127.0.0.1", port=8001)],
        ),
    ]
    protocol: Annotated[
        LogstashProtocol,
        Field(
            description="Protocol to communicate with logstash server.",
            default=LogstashProtocol.TCP,
        ),
    ]
    ssl_enabled: Annotated[
        bool, Field(description="Use TLS to communicate with logstash server.", default=True)
    ]
    ssl_verify: Annotated[
        bool,
        Field(
            description="Verify validity of TLS certificate when communicating with logstash.",
            default=True,
        ),
    ]


class GraylogConfig(BaseSchema):
    host: Annotated[str, Field(description="Graylog hostname.", examples=["127.0.0.1"])]
    port: Annotated[int, Field(description="Graylog server port number.", examples=[8000])]
    level: Annotated[LogLevel, Field(description="Log level.", default=LogLevel.INFO)]
    ssl_verify: Annotated[
        bool,
        Field(
            description="Verify validity of TLS certificate when communicating with logstash.",
            default=True,
        ),
    ]
    ca_certs: Annotated[
        str | None,
        Field(
            description="Path to Root CA certificate file.",
            examples=["/etc/ssl/ca.pem"],
            default=None,
        ),
    ]
    keyfile: Annotated[
        str | None,
        Field(
            description="Path to TLS private key file.",
            examples=["/etc/backend.ai/graylog/privkey.pem"],
            default=None,
        ),
    ]
    certfile: Annotated[
        str | None,
        Field(
            description="Path to TLS certificate file.",
            examples=["/etc/backend.ai/graylog/cert.pem"],
            default=None,
        ),
    ]


class PyroscopeConfig(BaseSchema):
    enabled: Annotated[bool, Field(default=False, description="Enable pyroscope profiler.")]
    app_name: Annotated[str, Field(default=None, description="Pyroscope app name.")]
    server_addr: Annotated[str, Field(default=None, description="Pyroscope server address.")]
    sample_rate: Annotated[int, Field(default=None, description="Pyroscope sample rate.")]


class LoggingConfig(BaseSchema):
    level: Annotated[LogLevel, Field(default=LogLevel.INFO, description="Log level.")]
    pkg_ns: Annotated[
        dict[str, LogLevel],
        Field(
            description="Override default log level for specific scope of package",
            default=default_pkg_ns,
        ),
    ]
    drivers: Annotated[
        list[LogDriver],
        Field(default=[LogDriver.CONSOLE], description="Array of log drivers to print."),
    ]
    console: Annotated[
        ConsoleLogConfig, Field(default=ConsoleLogConfig(colored=None, format=LogFormat.VERBOSE))
    ]
    file: Annotated[FileLogConfig | None, Field(default=None)]
    logstash: Annotated[LogstashConfig | None, Field(default=None)]
    graylog: Annotated[GraylogConfig | None, Field(default=None)]


class DebugConfig(BaseSchema):
    enabled: Annotated[bool, Field(default=False)]
    asyncio: Annotated[bool, Field(default=False)]
    enhanced_aiomonitor_task_info: Annotated[bool, Field(default=False)]
    log_events: Annotated[bool, Field(default=False)]


class WSProxyConfig(BaseSchema):
    ipc_base_path: Annotated[
        Path,
        Field(
            default=Path("/tmp/backend.ai/ipc"),
            description="Directory to store temporary UNIX sockets.",
        ),
    ]
    event_loop: Annotated[
        EventLoopType,
        Field(default=EventLoopType.ASYNCIO, description="Type of event loop to use."),
    ]
    pid_file: Annotated[
        Path,
        Field(
            default=Path(os.devnull),
            description="Place to store process PID.",
            examples=["/run/backend.ai/wsproxy/wsproxy.pid"],
        ),
    ]

    id: Annotated[
        str,
        Field(default=f"i-{socket.gethostname()}", examples=["i-node01"], description="Node id."),
    ]
    user: Annotated[
        int,
        UserID(default_uid=_file_perm.st_uid),
        Field(default=_file_perm.st_uid, description="Process owner."),
    ]
    group: Annotated[
        int,
        GroupID(default_gid=_file_perm.st_gid),
        Field(default=_file_perm.st_uid, description="Process group."),
    ]

    bind_host: Annotated[
        str,
        Field(
            default="0.0.0.0",
            description="Bind address of the port opened on behalf of wsproxy worker",
        ),
    ]
    advertised_host: Annotated[
        str, Field(examples=["example.com"], description="Hostname to be advertised to client")
    ]

    bind_api_port: Annotated[
        int, Field(default=5050, description="Port number to bind for API server")
    ]
    internal_api_port: Annotated[
        int, Field(default=15050, description="Port number to bind for internal API server")
    ]
    advertised_api_port: Annotated[
        int | None,
        Field(default=None, examples=[15050], description="API port number reachable from client"),
    ]

    bind_proxy_port_range: Annotated[
        tuple[int, int],
        Field(default=[10200, 10300], description="Port number to bind for actual traffic"),
    ]
    advertised_proxy_port_range: Annotated[
        tuple[int, int] | None,
        Field(
            default=None,
            examples=[[20200, 20300]],
            description="Traffic port range reachable from client",
        ),
    ]

    protocol: Annotated[
        ProxyProtocol, Field(default=ProxyProtocol.HTTP, description="Proxy protocol")
    ]

    jwt_encrypt_key: Annotated[
        str, Field(examples=["50M3G00DL00KING53CR3T"], description="JWT encryption key")
    ]
    permit_hash_key: Annotated[
        str, Field(examples=["50M3G00DL00KING53CR3T"], description="Permit hash key")
    ]

    api_secret: Annotated[str, Field(examples=["50M3G00DL00KING53CR3T"], description="API secret")]

    aiomonitor_termui_port: Annotated[
        int,
        Field(
            gt=0, lt=65536, description="Port number for aiomonitor termui server.", default=38500
        ),
    ]
    aiomonitor_webui_port: Annotated[
        int,
        Field(
            gt=0, lt=65536, description="Port number for aiomonitor webui server.", default=39500
        ),
    ]


class ServerConfig(BaseSchema):
    wsproxy: Annotated[WSProxyConfig, Field(default_factory=WSProxyConfig)]
    pyroscope: Annotated[PyroscopeConfig, Field(default_factory=PyroscopeConfig)]
    logging: Annotated[LoggingConfig, Field(default_factory=LoggingConfig)]
    debug: Annotated[DebugConfig, Field(default_factory=DebugConfig)]


def load(config_path: Path | None = None, log_level: LogLevel = LogLevel.NOTSET) -> ServerConfig:
    # Determine where to read configuration.
    raw_cfg, _ = config.read_from_file(config_path, "wsproxy")

    config.override_key(raw_cfg, ("debug", "enabled"), log_level == LogLevel.DEBUG)
    if log_level != LogLevel.NOTSET:
        config.override_key(raw_cfg, ("logging", "level"), log_level)
        config.override_key(raw_cfg, ("logging", "pkg-ns", "ai.backend"), log_level)
        config.override_key(raw_cfg, ("logging", "pkg-ns", "aiohttp"), log_level)

    # Validate and fill configurations
    # (allow_extra will make configs to be forward-copmatible)
    try:
        cfg = ServerConfig(**raw_cfg)
        if cfg.debug.enabled:
            print("== WSProxy configuration ==", file=sys.stderr)
            print(pformat(cfg.model_dump()), file=sys.stderr)
    except ValidationError as e:
        print(
            "ConfigurationError: Could not read or validate the wsproxy local config:",
            file=sys.stderr,
        )
        print(pformat(e.errors()), file=sys.stderr)
        raise click.Abort()
    else:
        return cfg


class Undefined:
    pass


class UnsupportedTypeError(RuntimeError):
    pass


def generate_example_json(
    schema: type[BaseSchema] | types.GenericAlias | types.UnionType, parent: list[str] = []
) -> dict | list:
    if isinstance(schema, types.UnionType):
        return generate_example_json(typing.get_args(schema)[0], parent=[*parent])
    elif isinstance(schema, types.GenericAlias):
        if typing.get_origin(schema) is not list:
            raise RuntimeError("GenericAlias other than list not supported!")
        return [generate_example_json(typing.get_args(schema)[0], parent=[*parent])]
    elif issubclass(schema, BaseSchema):
        res = {}
        for name, info in schema.model_fields.items():
            config_key = [*parent, name]
            assert info.annotation
            alternative_example = Undefined
            if info.examples:
                res[name] = info.examples[0]
            elif info.default != PydanticUndefined:
                alternative_example = info.default
            if name not in res:
                try:
                    res[name] = generate_example_json(info.annotation, parent=config_key)
                except RuntimeError:
                    if alternative_example != Undefined:
                        res[name] = alternative_example
                    else:
                        raise
        return res
    else:
        raise UnsupportedTypeError(str(schema))
